@@ -536,7 +536,7 @@ static int spuldrfs_fill_dir(struct dentry *dir,
 
 fail:
 
-	list_for_each_entry_safe(dentry, tmp, &dir->d_subdirs, d_u.d_child)
+	list_for_each_entry_safe(dentry, tmp, &dir->d_subdirs, d_child)
 		dput(dentry);
 
 	shrink_dcache_parent(dir);
