@@ -322,7 +322,7 @@ int aufs_unlink(struct inode *dir, struct dentry *dentry)
 	inode = dentry->d_inode;
 	IMustLock(inode);
 	err = -EISDIR;
-	if (unlikely(d_is_directory(dentry)))
+	if (unlikely(d_is_dir(dentry)))
 		goto out_unlock; /* possible? */
 
 	bstart = au_dbstart(dentry);
@@ -423,7 +423,7 @@ int aufs_rmdir(struct inode *dir, struct dentry *dentry)
 	inode = dentry->d_inode;
 	IMustLock(inode);
 	err = -ENOTDIR;
-	if (unlikely(!d_is_directory(dentry)))
+	if (unlikely(!d_is_dir(dentry)))
 		goto out_unlock; /* possible? */
 
 	err = -ENOMEM;
