@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2013 Junjiro R. Okajima
+ * Copyright (C) 2005-2015 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -137,7 +137,7 @@ struct file *au_xino_create2(struct file *base_file, struct file *copy_src)
 {
 	struct file *file;
 	struct dentry *base, *parent;
-	struct inode *dir;
+	struct inode *dir, *delegated;
 	struct qstr *name;
 	struct path path;
 	int err;
@@ -152,8 +152,8 @@ struct file *au_xino_create2(struct file *base_file, struct file *copy_src)
 	path.dentry = vfsub_lookup_one_len(name->name, parent, name->len);
 	if (IS_ERR(path.dentry)) {
 		file = (void *)path.dentry;
-		pr_err("%.*s lookup err %ld\n",
-		       AuLNPair(name), PTR_ERR(path.dentry));
+		pr_err("%pd lookup err %ld\n",
+		       base, PTR_ERR(path.dentry));
 		goto out;
 	}
 
@@ -161,7 +161,7 @@ struct file *au_xino_create2(struct file *base_file, struct file *copy_src)
 	err = vfs_create(dir, path.dentry, S_IRUGO | S_IWUGO, NULL);
 	if (unlikely(err)) {
 		file = ERR_PTR(err);
-		pr_err("%.*s create err %d\n", AuLNPair(name), err);
+		pr_err("%pd create err %d\n", base, err);
 		goto out_dput;
 	}
 
@@ -170,13 +170,19 @@ struct file *au_xino_create2(struct file *base_file, struct file *copy_src)
 				 O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE
 				 /* | __FMODE_NONOTIFY */);
 	if (IS_ERR(file)) {
-		pr_err("%.*s open err %ld\n", AuLNPair(name), PTR_ERR(file));
+		pr_err("%pd open err %ld\n", base, PTR_ERR(file));
 		goto out_dput;
 	}
 
-	err = vfsub_unlink(dir, &file->f_path, /*force*/0);
+	delegated = NULL;
+	err = vfsub_unlink(dir, &file->f_path, &delegated, /*force*/0);
+	if (unlikely(err == -EWOULDBLOCK)) {
+		pr_warn("cannot retry for NFSv4 delegation"
+			" for an internal unlink\n");
+		iput(delegated);
+	}
 	if (unlikely(err)) {
-		pr_err("%.*s unlink err %d\n", AuLNPair(name), err);
+		pr_err("%pd unlink err %d\n", base, err);
 		goto out_fput;
 	}
 
@@ -184,7 +190,7 @@ struct file *au_xino_create2(struct file *base_file, struct file *copy_src)
 		/* no one can touch copy_src xino */
 		err = au_copy_file(file, copy_src, vfsub_f_size_read(copy_src));
 		if (unlikely(err)) {
-			pr_err("%.*s copy err %d\n", AuLNPair(name), err);
+			pr_err("%pd copy err %d\n", base, err);
 			goto out_fput;
 		}
 	}
@@ -723,7 +729,7 @@ struct file *au_xino_create(struct super_block *sb, char *fname, int silent)
 {
 	struct file *file;
 	struct dentry *h_parent, *d;
-	struct inode *h_dir;
+	struct inode *h_dir, *inode;
 	int err;
 
 	/*
@@ -741,11 +747,16 @@ struct file *au_xino_create(struct super_block *sb, char *fname, int silent)
 	}
 
 	/* keep file count */
+	err = 0;
+	inode = file_inode(file);
 	h_parent = dget_parent(file->f_dentry);
 	h_dir = h_parent->d_inode;
 	mutex_lock_nested(&h_dir->i_mutex, AuLsc_I_PARENT);
 	/* mnt_want_write() is unnecessary here */
-	err = vfsub_unlink(h_dir, &file->f_path, /*force*/0);
+	/* no delegation since it is just created */
+	if (inode->i_nlink)
+		err = vfsub_unlink(h_dir, &file->f_path, /*delegated*/NULL,
+				   /*force*/0);
 	mutex_unlock(&h_dir->i_mutex);
 	dput(h_parent);
 	if (unlikely(err)) {
@@ -987,12 +998,10 @@ static au_readf_t find_readf(struct file *h_file)
 {
 	const struct file_operations *fop = h_file->f_op;
 
-	if (fop) {
-		if (fop->read)
-			return fop->read;
-		if (fop->aio_read)
-			return do_sync_read;
-	}
+	if (fop->read)
+		return fop->read;
+	if (fop->aio_read)
+		return do_sync_read;
 	return ERR_PTR(-ENOSYS);
 }
 
@@ -1000,12 +1009,10 @@ static au_writef_t find_writef(struct file *h_file)
 {
 	const struct file_operations *fop = h_file->f_op;
 
-	if (fop) {
-		if (fop->write)
-			return fop->write;
-		if (fop->aio_write)
-			return do_sync_write;
-	}
+	if (fop->write)
+		return fop->write;
+	if (fop->aio_write)
+		return do_sync_write;
 	return ERR_PTR(-ENOSYS);
 }
 
@@ -1197,9 +1204,7 @@ int au_xino_set(struct super_block *sb, struct au_opt_xino *xino, int remount)
 			cur_parent = dget_parent(cur_xino->f_dentry);
 			cur_name = &cur_xino->f_dentry->d_name;
 			skip = (cur_parent == parent
-				&& dname->len == cur_name->len
-				&& !memcmp(dname->name, cur_name->name,
-					   dname->len));
+				&& au_qstreq(dname, cur_name));
 			dput(cur_parent);
 		}
 		if (skip)
@@ -1221,6 +1226,8 @@ int au_xino_set(struct super_block *sb, struct au_opt_xino *xino, int remount)
 
 	/* reset all */
 	AuIOErr("failed creating xino(%d).\n", err);
+	au_xigen_clr(sb);
+	xino_clear_xib(sb);
 
 out:
 	dput(parent);
@@ -1296,10 +1303,9 @@ int au_xino_path(struct seq_file *seq, struct file *file)
 	int err;
 
 	err = au_seq_path(seq, &file->f_path);
-	if (unlikely(err < 0))
+	if (unlikely(err))
 		goto out;
 
-	err = 0;
 #define Deleted "\\040(deleted)"
 	seq->count -= sizeof(Deleted) - 1;
 	AuDebugOn(memcmp(seq->buf + seq->count, Deleted,
